import { toggleMark } from "prosemirror-commands";
import { TextSelection, } from "prosemirror-state";
import { Decoration, DecorationSet } from "prosemirror-view";
import { _t } from "../../shared/localization";
import { ManagedInterfaceKey, PluginInterfaceView, } from "../../shared/prosemirror-plugins/interface-manager";
import { StatefulPlugin } from "../../shared/prosemirror-plugins/plugin-extensions";
import { escapeHTML, generateRandomId, getPlatformModKey, } from "../../shared/utils";
/**
 * PluginView for @see linkEditorPlugin
 * @internal Exported for testing
 */
export class LinkEditor extends PluginInterfaceView {
    validateLink;
    viewContainer;
    get hrefInput() {
        return this.viewContainer.querySelector(".js-link-editor-href");
    }
    get textInput() {
        return this.viewContainer.querySelector(".js-link-editor-text");
    }
    get saveBtn() {
        return this.viewContainer.querySelector(".js-link-editor-save-btn");
    }
    get hrefError() {
        return this.viewContainer.querySelector(".js-link-editor-href-error");
    }
    constructor(view, validateLink) {
        super(LINK_EDITOR_KEY);
        this.validateLink = validateLink;
        const randomId = generateRandomId();
        this.viewContainer = document.createElement("form");
        this.viewContainer.className = "mt6 bt bb bc-black-400 js-link-editor";
        this.viewContainer.innerHTML = escapeHTML `<div class="d-flex fd-column gsy gs8 p12">
            <div class="flex--item">
                <label for="link-editor-href-input-${randomId}" class="s-label mb4">${_t("link_editor.href_label")}</label>
                <input id="link-editor-href-input-${randomId}" class="s-input js-link-editor-href" type="text" name="href" aria-describedby="link-editor-href-error-${randomId}" placeholder="https://"/>
                <p id="link-editor-href-error-${randomId}" class="s-input-message mt4 d-none js-link-editor-href-error"></p>
            </div>

            <div class="flex--item">
                <label for="link-editor-text-input-${randomId}" class="s-label mb4">${_t("link_editor.text_label")}</label>
                <input id="link-text-href-input-${randomId}" class="s-input js-link-editor-text" type="text" name="text" />
            </div>

            <div class="flex--item">
                <button class="s-btn s-btn__primary js-link-editor-save-btn" type="submit" disabled>${_t("link_editor.save_button")}</button>
                <button class="s-btn" type="reset">${_t("link_editor.cancel_button")}</button>
            </div>
        </div>`;
        this.viewContainer.addEventListener("submit", (e) => {
            e.preventDefault();
            e.stopPropagation();
            this.handleSave(view);
        });
        this.viewContainer.addEventListener("reset", (e) => {
            e.preventDefault();
            e.stopPropagation();
            let tr = this.tryHideInterfaceTr(view.state);
            tr = LINK_EDITOR_KEY.updateVisibility(false, tr);
            if (tr) {
                view.dispatch(tr);
            }
        });
        this.hrefInput.addEventListener("input", (e) => {
            this.validate(e.target.value);
        });
    }
    validate(href) {
        const valid = this.validateLink(href);
        if (!valid) {
            this.showValidationError(_t("link_editor.validation_error"));
        }
        else {
            this.hideValidationError();
        }
        this.saveBtn.disabled = !valid;
        return valid;
    }
    showValidationError(errorMessage) {
        const parent = this.hrefInput.parentElement;
        const error = this.hrefError;
        parent.classList.add("has-error");
        error.textContent = errorMessage;
        error.classList.remove("d-none");
    }
    hideValidationError() {
        const parent = this.hrefInput.parentElement;
        const error = this.hrefError;
        parent.classList.remove("has-error");
        error.textContent = "";
        error.classList.add("d-none");
    }
    resetEditor() {
        this.hrefInput.value = "";
        this.textInput.value = "";
        this.hideValidationError();
    }
    handleSave(view) {
        const href = this.hrefInput.value;
        if (!this.validate(href)) {
            return;
        }
        const text = this.textInput.value || href;
        const node = view.state.schema.text(text, []);
        let tr = this.tryHideInterfaceTr(view.state, {
            url: null,
            text: null,
        }) || view.state.tr;
        tr = LINK_EDITOR_KEY.updateVisibility(false, tr);
        // set the text first, inheriting all marks
        tr = tr.replaceSelectionWith(node, true);
        // reselect the now unselected text
        tr = tr.setSelection(TextSelection.create(tr.doc, tr.selection.from - text.length, tr.selection.to));
        // drop any link marks that might already exist
        tr = tr.removeMark(tr.selection.from, tr.selection.to, view.state.schema.marks.link);
        // add our link mark back onto the selection
        tr = tr.addMark(tr.selection.from, tr.selection.to, view.state.schema.marks.link.create({ href }));
        view.dispatch(tr);
    }
    update(view) {
        super.update(view);
        const state = LINK_EDITOR_KEY.getState(view.state);
        if (this.isShown) {
            if (state?.url) {
                this.hrefInput.value = state.url;
                this.validate(state.url);
            }
            if (state?.text) {
                this.textInput.value = state.text;
            }
            let tr = this.tryShowInterfaceTr(view.state);
            if (tr) {
                tr = LINK_EDITOR_KEY.forceHideTooltipTr(view.state.apply(tr));
                view.dispatch(tr);
            }
            this.hrefInput.focus();
        }
        else {
            this.resetEditor();
            let tr = this.tryHideInterfaceTr(view.state);
            if (tr) {
                tr = LINK_EDITOR_KEY.updateVisibility(false, tr);
                view.dispatch(tr);
            }
        }
    }
    destroy() {
        // this.uploadField.remove();
        this.viewContainer.remove();
    }
    buildInterface(container) {
        // add the view container to the menu area
        container.appendChild(this.viewContainer);
    }
    destroyInterface(container) {
        container.removeChild(this.viewContainer);
    }
}
/**
 * Custom PluginKey with additional methods for interacting with a LinkTooltip
 */
class LinkEditorPluginKey extends ManagedInterfaceKey {
    constructor() {
        super(LinkEditor.name);
    }
    /**
     * Force the link tooltip to hide - useful e.g. when the entire editor is losing
     * focus and we want to make sure the tooltip disappears, too
     */
    forceHideTooltipTr(state) {
        const meta = this.getState(state);
        // if the tooltip is not showing, just return
        if (!tooltipIsShowing(meta.decorations)) {
            return state.tr;
        }
        meta.forceHideTooltip = true;
        return this.setMeta(state.tr, meta);
    }
    updateVisibility(visibility, trans) {
        const meta = trans.getMeta(LINK_EDITOR_KEY);
        meta.visible = visibility;
        return this.setMeta(trans, meta);
    }
}
/** The plugin key the image uploader plugin is tied to */
export const LINK_EDITOR_KEY = new LinkEditorPluginKey();
/** Manages the decorations necessary for drawing the link editor tooltip */
class LinkTooltip {
    content;
    href;
    removeListener;
    editListener;
    get editButton() {
        return this.content.querySelector(".js-link-tooltip-edit");
    }
    get removeButton() {
        return this.content.querySelector(".js-link-tooltip-remove");
    }
    get link() {
        return this.content.querySelector("a");
    }
    constructor(state) {
        const popoverId = "link-tooltip-popover" + generateRandomId();
        this.content = document.createElement("span");
        this.content.className = "w0";
        this.content.setAttribute("aria-controls", popoverId);
        this.content.setAttribute("data-controller", "s-popover");
        this.content.setAttribute("data-s-popover-placement", "bottom");
        this.content.innerHTML = escapeHTML `<div class="s-popover is-visible p4 w-auto wmx-initial wmn-initial js-link-tooltip"
            id="${popoverId}"
            role="menu">
            <div class="s-popover--arrow"></div>
            <div class="d-flex ai-center">
                <a href="${this.href}"
                    class="wmx3 flex--item fs-body1 fw-normal truncate ml8 mr4 us-none"
                    target="_blank"
                    rel="nofollow noreferrer">${this.href}</a>
                <button type="button"
                        class="flex--item s-btn mr4 js-link-tooltip-edit"
                        title="${_t("link_tooltip.edit_button_title")}"><span class="svg-icon-bg iconPencilSm"></span></button>
                <button type="button"
                        class="flex--item s-btn js-link-tooltip-remove"
                        title="${_t("link_tooltip.remove_button_title")}"><span class="svg-icon-bg iconTrashSm"></span></button>
            </div>
        </div>`;
        // never allow the popover to hide itself. It either exists visibly or not at all
        this.content.addEventListener("s-popover:hide", (e) => {
            e.preventDefault();
        });
        // don't bind the exact listener, call whatever is currently set on `this` at event time
        const removeListener = () => {
            this.removeListener.call(this);
        };
        const editListener = () => {
            this.editListener.call(this);
        };
        // hook up the click/keyboard events for the supporting buttons
        this.bindElementInteraction(this.removeButton, removeListener);
        this.bindElementInteraction(this.editButton, editListener);
        this.update(state);
    }
    /**
     * Binds both a mousedown and selective keydown listener to replace the purposefully missing "click" event
     * @param element The element to bind the events to
     * @param callback The callback to run on mousedown/keydown
     */
    bindElementInteraction(element, callback) {
        element.addEventListener("mousedown", (e) => {
            e.stopPropagation();
            e.preventDefault();
            callback.call(this, e);
        });
        element.addEventListener("keydown", (e) => {
            // allow the Tab key to keep doing its thing
            if (e.key === "Tab") {
                return;
            }
            // cancel all other keypresses
            e.stopPropagation();
            e.preventDefault();
            // enter/space should still fire the event as if clicked
            if (e.key === "Enter" || e.key === " ") {
                callback.call(this, e);
            }
        });
    }
    /**
     * Updates the internal state / tooltip visuals based on the current editor state
     * @param state the current state of the editor
     */
    update(state) {
        if (!this.isLink(state)) {
            return;
        }
        const linkMarks = this.findMarksInSelection(state);
        if (linkMarks.length > 0) {
            this.href = linkMarks[0].attrs.href;
            const link = this.link;
            link.href = link.title = link.innerText = this.href;
        }
    }
    /**
     * Gets the tooltip decoration from a new PluginState.apply call
     * @param value The existing LinkTooltipState (with forceHideTooltip potentially set)
     * @param newState The state after the transaction
     */
    getDecorations(value, newState) {
        // if we're forced to hide the decorations, don't even attempt to create them
        if ("forceHideTooltip" in value && value.forceHideTooltip) {
            return DecorationSet.empty;
        }
        const marks = this.findMarksInSelection(newState);
        // if there are no marks in the current selection, then return empty
        if (!marks.length) {
            return DecorationSet.empty;
        }
        // Use the linkAround helper to determine the full range of the link mark.
        const linkRange = this.linkAround(newState);
        if (!linkRange) {
            return DecorationSet.empty;
        }
        // Only show the tooltip if the selection is entirely within the link.
        // If the selection starts before the link or ends after the link, do not show the tooltip.
        if (newState.selection.from < linkRange.from ||
            newState.selection.to > linkRange.to) {
            return DecorationSet.empty;
        }
        // always update the state, regardless of document changes (potential metadata changes can change tooltip visuals)
        this.update(newState);
        // Compute the midpoint of the link's range to anchor the tooltip.
        const pos = Math.floor((linkRange.from + linkRange.to) / 2);
        // create the widget tooltip via EditorView callback
        const decoration = Decoration.widget(pos, (view) => {
            /* NOTE: This function runs on every transaction update */
            this.updateEventListeners(view);
            return this.content;
        }, {
            // place the widget *before* the cursor so it isn't included in text selections
            side: -1,
            ignoreSelection: true,
            // cancel all events coming from inside this widget
            stopEvent: () => true,
        });
        return DecorationSet.create(newState.doc, [decoration]);
    }
    /**
     * Returns true if the focus event caused something in the content to be focused
     * @param e The dispatched focus event
     */
    hasFocus(e) {
        return this.content.contains(e.relatedTarget);
    }
    /**
     * Find out if the current selection contains a link mark
     * @param state The current editor state
     */
    isLink(state) {
        const { from, $from, to, empty } = state.selection;
        if (!empty) {
            return state.doc.rangeHasMark(from, to, state.schema.marks.link);
        }
        return (state.schema.marks.link.isInSet(state.storedMarks || $from.marks()) !== undefined);
    }
    /**
     * Expand the current selection to contain the entire link mark.
     * This allows us to remove the link mark from the entire link in the document
     * if the user didn't explicitly select a region to be toggled.
     */
    expandSelection(state, tr) {
        const expanded = this.linkAround(state);
        tr = tr.setSelection(TextSelection.create(tr.doc, expanded.from, expanded.to));
        return tr;
    }
    /**
     * Gets the positions immediately before and after a link mark in the current selection
     * @param state
     */
    linkAround(state) {
        const $pos = state.selection.$from;
        const start = $pos.parent.childAfter($pos.parentOffset);
        if (!start.node) {
            return;
        }
        const link = findMarksInSet(start.node.marks, state.schema.marks.link)[0];
        if (!link) {
            return;
        }
        let startIndex = $pos.index();
        let startPos = $pos.start() + start.offset;
        while (startIndex > 0 &&
            link.isInSet($pos.parent.child(startIndex - 1).marks)) {
            startIndex -= 1;
            startPos -= $pos.parent.child(startIndex).nodeSize;
        }
        let endIndex = $pos.indexAfter();
        let endPos = startPos + start.node.nodeSize;
        // Don't double-count the length of the the node if we're at the start of it
        if (endIndex !== $pos.index()) {
            while (endIndex < $pos.parent.childCount &&
                link.isInSet($pos.parent.child(endIndex).marks)) {
                endPos += $pos.parent.child(endIndex).nodeSize;
                endIndex += 1;
            }
        }
        return { from: startPos, to: endPos };
    }
    /**
     * Finds all marks in the current selection
     * @param state The current editor state
     */
    findMarksInSelection(state) {
        const linkMarks = [];
        const { to, from, $from, empty } = state.selection;
        if (empty) {
            return findMarksInSet($from.marks(), state.schema.marks.link);
        }
        if (to > from) {
            state.doc.nodesBetween(from, to, (node) => {
                linkMarks.push(findMarksInSet(node.marks, state.schema.marks.link));
            });
        }
        const returnValue = [];
        return returnValue.concat(...linkMarks);
    }
    /**
     * Updates apply/delete button events with the current editor view
     * @param view The current editor view
     */
    updateEventListeners(view) {
        this.removeListener = () => {
            let state = view.state;
            if (view.state.selection.empty) {
                const tr = this.expandSelection(state, state.tr);
                state = view.state.apply(tr);
            }
            toggleMark(view.state.schema.marks.link)(state, view.dispatch.bind(view));
        };
        this.editListener = () => {
            let tr = view.state.tr;
            tr = this.expandSelection(view.state, view.state.tr);
            const state = view.state.apply(tr);
            const { from, to } = state.selection;
            const text = state.doc.textBetween(from, to);
            const href = this.findMarksInSelection(state)[0].attrs
                .href;
            tr =
                LINK_EDITOR_KEY.showInterfaceTr(state, {
                    forceHideTooltip: true,
                    url: href,
                    text,
                }) || tr;
            view.dispatch(tr);
        };
    }
}
function tooltipIsShowing(linkDecos) {
    return linkDecos && linkDecos !== DecorationSet.empty;
}
/**
 * A plugin view that shows a tooltip when selecting a link in rich-text mode.
 * The tooltip shows the href attribute of the selected link and allows removing
 * the link mark from the document. Clicking on the tooltip's edit button will launch
 * a plugin view that allows editing the link's href and text.
 *
 * Note: This is not a _NodeView_ because when dealing with links, we're dealing with
 * _marks_, not _nodes_.
 */
export const linkEditorPlugin = (features) => new StatefulPlugin({
    key: LINK_EDITOR_KEY,
    state: {
        init(_, instance) {
            return {
                visible: false,
                linkTooltip: new LinkTooltip(instance),
                decorations: DecorationSet.empty,
                shouldShow: false,
            };
        },
        apply(tr, value, state, newState) {
            // check if force hide was set and add to value for getDecorations to use
            const meta = this.getMeta(tr) || value;
            if ("forceHideTooltip" in meta) {
                value.forceHideTooltip = meta.forceHideTooltip;
            }
            // update the linkTooltip and get the decorations
            let decorations = value.linkTooltip.getDecorations(value, newState);
            //get decoration for current selection
            const sel = state.selection;
            let selectionDecoration = null;
            if (meta.visible && sel && sel.from !== sel.to) {
                selectionDecoration = Decoration.inline(sel.from, sel.to, {
                    class: "bg-black-225",
                });
                decorations = decorations.add(newState.doc, [
                    selectionDecoration,
                ]);
            }
            return {
                ...meta,
                forceHideTooltip: value.forceHideTooltip && tooltipIsShowing(decorations),
                linkTooltip: value.linkTooltip,
                decorations: decorations,
            };
        },
    },
    props: {
        decorations(state) {
            return this.getState(state).decorations || DecorationSet.empty;
        },
        handleDOMEvents: {
            /** Handle editor blur and close the tooltip if it isn't focused */
            blur(view, e) {
                const { linkTooltip, decorations } = LINK_EDITOR_KEY.getState(view.state);
                // if the editor blurs, but NOT because of the tooltip, hide the tooltip
                if (tooltipIsShowing(decorations) &&
                    !view.hasFocus() &&
                    !linkTooltip.hasFocus(e)) {
                    view.dispatch(LINK_EDITOR_KEY.forceHideTooltipTr(view.state));
                }
                // always return false since we're not cancelling/handling the blur
                return false;
            },
        },
        /** Handle mod-click to open links in a new window */
        handleClick(view, pos, event) {
            const mark = findMarksInSet(view.state.doc.resolve(pos).marks(), view.state.schema.marks.link)[0];
            const modPressed = getPlatformModKey() === "Cmd"
                ? event.metaKey
                : event.ctrlKey;
            const handled = mark && modPressed;
            // a link was mod-clicked, so open it in a new window
            if (handled) {
                window.open(mark.attrs.href, "_blank");
            }
            return !!handled;
        },
    },
    view(editorView) {
        return new LinkEditor(editorView, features.validateLink);
    },
});
/**
 * Dispatches a transaction to show the link editor with the given url and text filled in
 * @param view The current editor view
 * @param url The value to prefill the url input with
 * @param text The value to prefill the text input with
 */
export function showLinkEditor(view, url, text) {
    let tr = LINK_EDITOR_KEY.showInterfaceTr(view.state, {
        url,
        text,
    });
    tr = LINK_EDITOR_KEY.updateVisibility(true, tr);
    if (tr) {
        view.dispatch(tr);
    }
}
/**
 * Dispatches a transaction to hide the link editor
 * @param view The current editor view
 */
export function hideLinkEditor(view) {
    let tr = LINK_EDITOR_KEY.hideInterfaceTr(view.state, {
        url: null,
        text: null,
    });
    tr = LINK_EDITOR_KEY.updateVisibility(false, tr);
    if (tr) {
        view.dispatch(tr);
    }
}
/**
 * Finds all marks in a set that are of the given type
 * @param marks The set of marks to search
 * @param type The type of mark to find
 */
function findMarksInSet(marks, type) {
    return marks.filter((mark) => mark.type === type);
}
